import numpy as np
import pandas as pd
import json
from .consts import eps0, hbar, h, c
from sympy.physics.wigner import wigner_6j
from .atom import Atom
from .auxiliary import _tex_state


class LightshiftSolver():
    def __init__(self, atom_filename, transitions_filename, Fi=None):
        """
        Calculate dynamical scalar, vector and tensor light shifts
        for atomic states in the presence of hyperfine coupling.

        Atom states and transitions json files need to be provided, see
        README.md.

        Arguments:
            atom_filename (str): path to atom states json file
            transitions_filename (str): path to transitions json file
            Fi (float): quantum number F of intitial state (default F=I if None)

        Returns:
            lightshift_solver: obect wrapping lightshift solver methods
        """
        self.atom, self.transitions = self._import_states_and_transitions(atom_filename, transitions_filename)
        self.states = self.atom.states
        self.I = self.atom.I
        if Fi is None:
            self.Fi = self.I
        else:
            self.Fi = Fi
        self.state_i = self.transitions[0]['state_i']

    def _import_states_and_transitions(self, atom_filename, transitions_filename):
        try:
            a = Atom.from_json(atom_filename)
        except Exception as e:
            raise Exception('Atom file not found or not in json format.', e)
        try:
            t = self._import_json(transitions_filename)
        except:
            raise Exception('Transitions file not found or not in json format.')

        return a, t

    @staticmethod
    def _import_json(fn):
        with open(fn, 'r') as f:
                content = json.load(f)
        return content

    def _calc_Fs(self, J):
        a = np.abs(J-self.I)
        b = J + self.I
        r = int(b-a)+1

        return [a + f for f in range(r)]

    def _query_state_prop(self, state, prop):
        configuration, term = state[0], state[1]
        return self.states[configuration][term][prop]

    def _query_state_prop_exist(self, state, prop):
        configuration, term = state[0], state[1]
        return prop in self.states[configuration][term]

    def final_states(self):
        return [t['state_f'] for t in self.transitions]

    def frequency(self, state):
        return self._query_state_prop(state, 'frequency')

    def transition_frequency(self, state_f):
        """
        calculate transition frequency between initial state and state_f
        """
        fi = self.frequency(self.state_i)
        ff = self.frequency(state_f)
        return (ff-fi)

    def transition_frequency_hyperfine(self, state_f, Ff):
        """
        calculate transition frequency between initial state (F=self.Fi) and
        state_f (F=Ff), including hyperfine shift
        """
        hfs_i = self.hyperfine_shift(self.state_i, self.Fi)
        hfs_f = self.hyperfine_shift(state_f, Ff)

        fi = self.frequency(self.state_i) + hfs_i
        ff = self.frequency(state_f) + hfs_f
        return (ff-fi)

    def sorted_transitions_dict(self, hyperfine=False):
        d = []
        if hyperfine:
            for s in self.final_states():
                for F in self._calc_Fs(self._J_from_state(s)):
                    d.append({
                        'state':s,
                        'F': F,
                        'frequency': self.transition_frequency_hyperfine(s,F)})
        else:
            for t in self.transitions:
                    state_f = t['state_f']
                    state_i = self.state_i
                    states = sorted([state_i, state_f], key=self.frequency)
                    state_l = states[0]
                    state_u = states[1]
                    br =  self.atom.branching_ratio_LS(tuple(state_u),tuple(state_l))
                    if np.isnan(br):
                        lt_f = 1/t['Gamma']
                    else:
                        lt_f =  1/(t['Gamma'] / br)

                    d.append({
                        'state_i': self.state_i,
                        'state_f': state_f,
                        'frequency': self.transition_frequency(state_f),
                        '_ref_frequency': self._query_state_prop(state_u,
                            '_ref_frequency'),
                        'wavenumber': self.transition_frequency(state_f)/c/100,
                        'Gamma': t['Gamma'],
                        '_ref_Gamma': t['_ref_Gamma'],
                        'branching_ratio_LS': br,
                        'lifetime_f': lt_f
                        })
        sd = sorted(d, key=lambda x: x['frequency'])
        for s in sd:
            s['wavelength'] = c/s['frequency']
        return sd

    def pretty_transition_table(self, latex=False):
        sorted_transitions_dict = self.sorted_transitions_dict()
        trans = pd.DataFrame(sorted_transitions_dict)
        trans[r'final state']=trans['state_f'].apply(_tex_state)
        trans[r'$\lambda$ (nm)'] = trans['wavelength'].apply(lambda x: '{:1.3f}'.format(x*1e9))
        trans[r'k (1/cm)'] = trans['wavenumber'].apply(lambda x: '{:1.2f}'.format(x))
        trans[r'$\Gamma$ (MHz)'] = trans['Gamma'].apply(lambda x: '{:1.2f}'.format(x/1e6))
        trans[r'$\tau_f$ (ns)'] = trans['lifetime_f'].apply(lambda x: '{:1.3f}'.format(x*1e9))
        trans['br.r. (LS)'] = trans['branching_ratio_LS'].apply(\
                lambda x: 'n/a' if np.isnan(x) else '{:1.4f}'.format(x))
        trans[r'ref. ($\tau_f$, $\Gamma$)'] = trans['_ref_Gamma'].apply(lambda x: '{:s}'.format(x))
        trans_export = trans[['final state', r'$\lambda$ (nm)', 'k (1/cm)', r'$\tau_f$ (ns)',
                              'br.r. (LS)', r'$\Gamma$ (MHz)', r'ref. ($\tau_f$, $\Gamma$)']]

        if latex: 
            return trans_export.to_latex(index=False, escape=False,
                float_format=True, na_rep='n/a', column_format='lrrrrrl')
        else: 
            return trans_export

    @classmethod
    def finestruct_reduced_mat_el_sqd(cls, Ji, Jf, omega_Jf_Ji, Gamma):
        # reduced fine structure dipole element <Ji |d| Jf>
        # Steck 7.242, 7.296, gamma must be decay rate from higher to lower state,
        # omega is omega_f - omega_i
        if omega_Jf_Ji > 0:
            # initial state is lower in energy
            Je = Jf
            Jg = Ji
            conj_ratio = 1
        else:
            # initial state is higher in energy
            Je = Ji
            Jg = Jf
            conj_ratio = (2*Jg+1)/(2*Je+1) # see Steck 7.250, conj. red. mat. el.

        fma_eg = Gamma*3*np.pi*eps0*hbar*c**3/(abs(omega_Jf_Ji)**3)\
                             *(2*Je+1)/(2*Jg+1)

        return fma_eg * conj_ratio

    @classmethod
    def hyperfine_reduced_mat_el_sqd(cls, I, Fi, Ff, Ji, Jf, omega_Jf_Ji, Gamma):
        fma = cls.finestruct_reduced_mat_el_sqd(Ji, Jf, omega_Jf_Ji, Gamma)
        y = (2*Ji+1) * (2*Ff+1)
        z = (float(wigner_6j(float(Ji), float(Jf), 1, float(Ff), float(Fi), float(I))))**2
        return fma * y * z

    def _J_from_state(self, state):
        config, term = state
        try:
            J = self.states[config][term]['J']
            return J
        except:
            if len(term)==3:
                S, L, J = term
                J = int(J)
                return J
            else:
                raise Exception('State (%s, %s) has no J provided.'%(config,term))

    def scalar_polarizability(self, lam):
        """
        Calculate scalar polarizability in SI units for initial state (self.state_i) for
        light with a given wavelength lam in nm.

        Arguments:
            lam (float): wavelength in nm
        Returns:
            float: scalar polarizability in SI units
        """
        omega = 2*np.pi*c/lam

        p = 0.
        for trans in self.transitions:
            Ji = self._J_from_state(trans['state_i'])
            Jf = self._J_from_state(trans['state_f'])
            omega_Ff_Fi = 2*np.pi*self.transition_frequency(trans['state_f'])
            Gamma = trans['Gamma']

            for Ff in self._calc_Fs(Jf):
                a = 2*omega_Ff_Fi*self.hyperfine_reduced_mat_el_sqd(self.I, self.Fi, Ff, Ji, Jf,
                                                           omega_Ff_Fi, Gamma)
                b = 3*hbar*(omega_Ff_Fi**2 - omega**2)

                p = p + a/b

        return p

    def scalar_lightshift(self, lam, laser_intensity=1):
        """
        lightshift in Hz for lambda in nm and laser_intensity in W/cm^2
        """
        laser_intensity_unit = 1/(0.01)**2  # normalization Watt/cm^2
        p = self.scalar_polarizability(lam)
        ls = -p/(2*eps0*c)/h*laser_intensity_unit*laser_intensity
        return ls

    @staticmethod
    def _hyperfine_shift(F, I, J, hyper_A, hyper_B):
        """
        hyperfine coupling shift in Hz,
        for hyper_A and hyper_B in Hz.
        see Steck (7.134)
        A. Lurio, M. Mandel, and R. Novick. Second-Order Hyperfine and Zeeman Corrections
        for an sl-Configuration. Physical Review, 126(5):1758–1767, June 1962.
        """
        def K(F):
            return F * (F + 1) - I * (I + 1) - J * (J + 1)

        # dipole hyperfine coupling dEa
        dEa = K(F) * hyper_A/2

        # quadrupole hyperfine coupling dEb
        D = 2*I * (2*I - 1) * 2*J * (2*J - 1)
        if D == 0:
            dEb = 0
        else:
            dEb = (3 * K(F) * (K(F)+1) / 2 - 2*I * (I+1) * J * (J+1)) / D * hyper_B
        return (dEa + dEb)

    def hyperfine_shift(self, state, F):
        J = self._J_from_state(state)

        if self._query_state_prop_exist(state, 'hyper_A'):
            hyper_A = self._query_state_prop(state, 'hyper_A')
        else:
            hyper_A = 0
        if self._query_state_prop_exist(state, 'hyper_B'):
            hyper_B = self._query_state_prop(state, 'hyper_B')
        else:
            hyper_B = 0

        return self._hyperfine_shift(F, self.I, J, hyper_A, hyper_B)


    def polarizabilities(self, lam):
        """
        Calculate scalar, vector and tensor polarizability in SI units
        for initial state (self.state_i) for
        light with a given wavelength lam in nm.

        Arguments:
            lam (float): wavelength in nm
        Returns:
            (float,float,float) tuple: scalar, vector, tensor polarizability in SI units
        """
        omega = 2*np.pi*c/lam
        Fi = self.Fi

        p0, p1, p2 = (0,0,0)

        for trans in self.transitions:
            state_f = trans['state_f']
            Ji = self._J_from_state(trans['state_i'])
            Jf = self._J_from_state(trans['state_f'])
            Gamma = trans['Gamma']

            for Ff in self._calc_Fs(Jf):
                omega_Ff_Fi = 2*np.pi*self.transition_frequency_hyperfine(state_f, Ff)
                matel = self.hyperfine_reduced_mat_el_sqd(self.I, self.Fi, Ff, Ji, Jf,
                                                 omega_Ff_Fi, Gamma)
                f = omega_Ff_Fi*matel/(hbar*(omega_Ff_Fi**2 - omega**2))

                p0 += 2/3 * f
                p1 += (-1)**(Fi+Ff+1) * np.sqrt(6*Fi*(2*Fi+1)/(Fi+1))\
                      * float(wigner_6j(1,1,1,float(Fi),float(Fi),float(Ff))) * f*(omega/omega_Ff_Fi)
                p2 += (-1)**(Fi+Ff) * np.sqrt(40*Fi*(2*Fi+1)*(2*Fi-1)/(3*(Fi+1)*(2*Fi+3)))\
                      * float(wigner_6j(1,1,2,float(Fi),float(Fi),float(Ff))) * f

        return p0,p1,p2


    def lightshifts(self, lam, q, mFi, laser_intensity=1.):
        """
        Calculate scalar, vector and tensor light shift in Hz
        for initial state (self.state_i) with m_F = mFi for
        light with a given wavelength lam in nm and polarization q.

        Arguments:
            lam (float): wavelength in nm
            q (int): -1,0,1 for sigma-, pi, sigma+ polarization
            mFi: m_F component of the light shifted state
            laser_intensity (float): light intensity in W/cm^2
        Returns:
            (float,float,float) tuple: scalar, vector, tensor light shift in Hz
        """
        laser_intensity_unit = 1/(0.01)**2  # normalization Watt/cm^2

        Fi = self.Fi
        p0,p1,p2 = self.polarizabilities(lam)

        E02 = laser_intensity*laser_intensity_unit/(2*eps0*c)/h

        # scalar light shift
        l0 = -p0*E02
        if Fi > 0:
            l1 = -p1*mFi/Fi*(1)*q*E02
        else:
            l1 = 0
        if Fi > 1/2:
            l2 = -p2*(3*mFi**2-Fi*(Fi+1))/(Fi*(2*Fi-1))*(3*(1-np.abs(q))*E02-E02)/2
        else:
            l2 = 0
            

        return l0,l1,l2

    def total_lightshift(self, lam, q, mFi, laser_intensity=1.):
        """
        Calculate total scalar, vector and tensor light shift (summed) in Hz
        for initial state (self.state_i) with m_F = mFi for
        light with a given wavelength lam in nm and polarization q.

        Arguments:
            lam (float): wavelength in nm
            q (int): -1,0,1 for sigma-, pi, sigma+ polarization
            mFi: m_F component of the light shifted state
            laser_intensity (float): light intensity in W/cm^2
        Returns:
            float: total light shift (scalar, vector, tensor summed) in Hz
        """
        l0,l1,l2 = self.lightshifts(lam, q, mFi, laser_intensity=laser_intensity)
        return l0+l1+l2



from pytest import fixture
import sys
import os

here = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(here, '..'))


@fixture
def fixtures_path():
    return os.path.abspath(os.path.join(here, 'fixtures'))


@fixture
def atom_file(fixtures_path):
    return os.path.join(fixtures_path, 'atom_yb173.json')


@fixture
def trans_file(fixtures_path):
    return os.path.join(fixtures_path, 'transitions_1S0.json')



import pytest

from lightshifts import Atom


def test_branching_ratio_LS(atom_file):
    yb = Atom.from_json(atom_file)
    state_i = ('6s8s','3S1')
    state_f = ('6s6p','3P2')
    br = yb.branching_ratio_LS(state_i, state_f)
    assert pytest.approx(0.49202592529614686)==float(br)



import numpy as np
import json
from sympy.physics.wigner import wigner_6j


class Atom():
    def __init__(self, atom_dict):
        self.atom_dict = atom_dict
        self.states = self.atom_dict['states']
        self.I = self.atom_dict['I']

        # flatten hirarchy of states dict
        self.states_flat = {}
        for s0 in self.states:
            for s1 in self.states[s0]:
                self.states_flat[(s0,s1)] = self.states[s0][s1]

    @classmethod
    def from_json(cls, atom_filename):
        atom_dict = cls._import_states_from_json(atom_filename)
        return cls(atom_dict)

    @classmethod
    def _import_states_from_json(cls, atom_filename):
        try:
            atom_dict = cls._import_json(atom_filename)
        except:
            raise Exception('Atom file not found or not in json format.')
        return atom_dict

    @staticmethod
    def _import_json(fn):
        with open(fn, 'r') as f:
                content = json.load(f)
        return content


    @staticmethod
    def _parse_term_SLJ(term):
        S, L, J = term
        S = (int(S) - 1) / 2
        L = ('S', 'P', 'D', 'F', 'G').index(L)
        J = int(J)

        return S, L, J

    @classmethod
    def _dipole_coeff(cls, state_i, state_f, parity_selection=True):
        """fine-structure reduced matrix element, Steck 7.274
        modified with additional delta function for unchanged spin
        and selection rule on wave function parity:
        delta_l (small) = +-1 (parity must change)
        """
        S, L, J = cls._parse_term_SLJ(state_i[1])
        Sf, Lf, Jf = cls._parse_term_SLJ(state_f[1])

        if parity_selection:
            if cls.parity(state_i) == cls.parity(state_f): 
                return 0
        if S != Sf:
            return 0
        A = (-1)**(Jf + L + 1 + S)
        B = np.sqrt((2*Jf + 1) * (2*L + 1))
        C = wigner_6j(L, Lf, 1, Jf, J, S)

        return A*B*C

    @classmethod
    def parity(cls, state):
        """
        get electron configuration parity for two electron system
        used for selection rule for calculating branching ratio.
        """
        def one_electron_parity(t):
            if (t in ('s','d','g')):
                return 1
            elif (t in ('p','f')):
                return -1
            else: 
                return '?'
        try:
            orbitals = [s for s in state[0]][1::2]
            parities = [one_electron_parity(o) for o in orbitals]
            if len(orbitals)==0:
                return '?'
            return np.prod(parities)
        except:
            return '?'

    def frequency(self, state):
        """
        Frequency of state in Hz
        """
        configuration, term = state[0], state[1]
        return self.states[configuration][term]['frequency']


    def transition_omega(self, state_i, state_f):
        """
        calculate transition angular frequency
        """
        fi = self.frequency(state_i)
        ff = self.frequency(state_f)
        return 2*np.pi * (ff-fi)

    def states_below(self, state):
        """give a dictionary of all states lower in energy than state"""
        return {k: v for k, v in self.states_flat.items() \
                if v['frequency']<self.frequency(state)}

    def states_above(self, state):
        """give a dictionary of all states larger in energy than state"""
        return {k: v for k, v in self.states_flat.items() \
                if v['frequency']>self.frequency(state)}


    def branching_ratios_LS_dict(self, state_i, parity_selection=True, verbose=False):
        """
        Estimate branching ratios between LS coupling states. Gives branching
        ratio of an initial state state_i into all energetically lower lying states
        in the imported state library.
        Useful when only the lifetime of a state is known and
        not the decay rates into final states.
        Estimates the ratio of dipole matrix elements between a slection of states,
        using the parity selection rule (e.g. 6s6s -> 6s6p is allowed)
        for the electron configuration and angular momentum selection.

        Arguments:
            state_i (config,term)-tuple: initial state that branches into final states
            parity_selection (bool): use parity selection?
            verbose (bool): give more printed output
        Returns:
            dict: of (state: branching ratio) entries
        """
        # all the states energetically below state_i
        fs = list((s for s in self.states_below(state_i)))
        ratios = np.zeros(len(fs))

        for j, f in enumerate(fs):
            omega = self.transition_omega(state_i, f)
            try:
                M = self._dipole_coeff(state_i, f, parity_selection=parity_selection)
                ratios[j] = np.abs(omega**3 * M**2)
            except:
                pass

        if np.sum(ratios)==0:
            branching_ratios = np.array([np.nan for r in ratios])
        else:
            branching_ratios = ratios/np.sum(ratios)
        if verbose:
            for j, f in enumerate(fs):
                b = branching_ratios[j]
                if b>0: 
                    print('branching ratio into ', f, '=', b)
        return {f:float(b) for (f,b) in zip(fs, branching_ratios)}

    def branching_ratio_LS(self, state_i, state_f, parity_selection=True, verbose=False):
        """
        Estimate branching ratio between LS coupling state state_i and all energetically
        lower lying states in the imported state library.
        Useful when only the lifetime of a state is known and
        not the decay rates into final states.
        Estimates the ratio of dipole matrix elements between a slection of states,
        using the parity selection rule (e.g. 6s6s -> 6s6p is allowed)
        for the electron configuration and angular momentum selection.

        Arguments:
            state_i (config,term)-tuple: initial state that branches into state_f
            state_f (config,term)-tuple: final state
            parity_selection (bool): use parity selection?
            verbose (bool): give more printed output
        Returns:
            float: branching ratio
        """

        ratios = self.branching_ratios_LS_dict(state_i, parity_selection=parity_selection,
                verbose=verbose)
        br = ratios[state_f]
        if verbose:
            print('%s -> %s: %1.4f'%(state_i, state_f, br))
        return br



kB = 1.3806488e-23      # boltzmann constant
u = 1.66053892e-27      # atomic mass unit (proton mass)
h = 6.626070040e-34     # planck h
hbar = 1.054571800e-34  # planck hbar
c = 299792458           # speed of light
e = 1.602176565e-19     # electron charge
me = 9.10938291e-31     # electron mass
a0 = 5.291772109e-11    # bohr radius
muB = 9.274009994e-24   # bohr magneton
eps0 = 8.854187817e-12  # vacuum permittivity



import pytest
import numpy as np
from lightshifts.auxiliary import smart_gen_array, laser_intensity, frequency_from_wavenumber


def test_smart_gen_array():
    xx, yy = smart_gen_array(lambda x: x, 1, 2, sing=[1.5], n=5, eps=1e-2)
    print(xx,yy)
    assert (len(xx)==12)

def test_laser_intensity():
    li = laser_intensity(1, 1e-2)
    assert li == pytest.approx(2/np.pi)

def test_frequency_from_wavenumber():
    f = frequency_from_wavenumber(1)
    print(f)
    assert f == pytest.approx(29979245800)



import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np

from lightshifts import Atom
from lightshifts.consts import c


def smart_gen_array(func, a, b, sing=[], n=100, eps=1e-17):
    """
    Evaluate function func between a and b and replace singularities
    with np.nan. Every subinterval between the singularities is filled
    with n points and we evaluate up to eps around the singularity.
    """
    # make intervals between singularities
    intervals = []
    sorted_sing = sorted(sing)

    if len(sorted_sing) == 0:
        intervals.append([a, b])
    else:
        for i, s in enumerate(sorted_sing):
            if s < a:
                continue

            # first interval
            if s > a and len(intervals) == 0:
                intervals.append([a, s])

            # last interval
            elif s > b:
                intervals.append([sorted_sing[i-1], b])

            # intermediate intervals
            else:
                intervals.append([sorted_sing[i-1], s])

    if (intervals[-1][1] < b):
        intervals.append([intervals[-1][1], b])

    # solve function within the intervals and stitch with np.nan
    # on the singularities

    xx = []
    yy = []

    for interval in intervals:
        a, b = interval

        # safety margin
        a += eps
        b -= eps

        x = np.linspace(a, b, n)
        y = func(x)

        x = np.append(x, b)
        y = np.append(y, np.nan)

        xx = np.append(xx, x)
        yy = np.append(yy, y)

    return np.array(xx), np.array(yy)


def laser_intensity(laser_power, beam_waist):
    """
    Laser intensity in Watts/cm^2 for laser power in W and and beam waist in m
    """
    return 2/np.pi * laser_power/(beam_waist/1e-2)**2

def _tex_state(state, F=None, show_term=True):
    term, config = state
    try:
        par = Atom.parity(state)
        if par=='?':
            return '%s %s'%(term,config)
    except:
        pass
    if F is None:
        Fstr = ""
    else:
        if (F*2)%2==0:
            Fstr = " (F=%d)"%F
        else:
            Fstr = " (F=%d/2)"%(F*2)

    if show_term:
        term_str = '(%s)'%term
    else:
        term_str = ''
    return term_str+r"${}^{%s}\mathrm{%s}_{%s}$"%(config[0], config[1],
            config[2])+Fstr



def plot_total_lightshift_around_hyperfine_state(atom_filename, transitions_filename,
                                       state_f, Ff, q,
                                       Fi=None, df_min=10e9, df_max=10e9, n=200):
    
    from lightshifts import LightshiftSolver

    # make the plot data
    ls = LightshiftSolver(atom_filename, transitions_filename, Fi=Fi)
    fzero = ls.transition_frequency_hyperfine(state_f=state_f, Ff=Ff)
    lam_0 = c/(fzero+df_max)
    lam_1 = c/(fzero-df_min)
    sing = [x['wavelength'] for x in ls.sorted_transitions_dict(hyperfine=True)]
    if q!=0:
        possible_mf = np.arange(-ls.Fi, ls.Fi+1, 1)
    else:
        possible_mf = np.arange(ls.Fi, 0, -1)
    ls_data = []
    for mf in possible_mf:
        ls_data.append(
            smart_gen_array(lambda w: ls.total_lightshift(w, q, mf),
                            lam_0, lam_1, sing, n=n))

    # plot title
    title = {
        -1: '$\sigma^-$ light',
        1: '$\sigma^+$ light',
        0: '$\pi$ light'
    }
    # legend title
    leg_title = {
        -1: '$m_F$',
        1: '$m_F$',
        0: '$|m_F|$'
    }

    # create color map to identify mf component
    fmax = ls.Fi
    norm = mpl.colors.Normalize(vmin=-fmax, vmax=fmax)
    cmap = cm.copper
    m = cm.ScalarMappable(norm=norm, cmap=cmap)

    for i, mf in enumerate(possible_mf):
        w, l = ls_data[i]
        plt.plot((c/w-fzero)/1e9, l, c=m.to_rgba(mf), label=r'{:+1.0f}/2'.format(mf*2))

    for w in sing:
        plt.plot([(c/w-fzero)/1e9, (c/w-fzero)/1e9], [-1e6, 1e6], ':', color='gray', label='')

    plt.gca().axhline(y=0, color='k', lw=0.5)
    plt.xlim(-df_min/1e9,df_max/1e9)
    plt.xlabel(r'detuning from %s $\rightarrow$ %s (GHz)'%(_tex_state(ls.state_i),
                                                           _tex_state(state_f,
                                                               Ff)))

    plt.ylabel(r'$\Delta V_\mathrm{ac}/I_0$ [Hz/($W/\mathrm{cm}^2$)]')
    plt.legend(loc='upper left', title=leg_title[q], frameon=False)
    plt.title(title[q])


def plot_scalar_lightshift(atom_filename, transitions_filename,
              lam_min=200e-9, lam_max=1500e-9, n=200):

    from lightshifts import LightshiftSolver

    # make the plot data
    ls = LightshiftSolver(atom_filename, transitions_filename)
    lam_0 = lam_min
    lam_1 = lam_max
    sing = [x['wavelength'] for x in ls.sorted_transitions_dict()]
    w, l = smart_gen_array(lambda w: ls.scalar_lightshift(w), lam_0, lam_1, sing, n=n)

    plt.plot(w*1e9, l, label='(%s)%s'%(ls.state_i[0],ls.state_i[1]))

    for w in sing:
        plt.plot([w*1e9, w*1e9], [-1e6, 1e6], ':', color='gray', label='')
    plt.gca().axhline(y=0, color='k', lw=0.5)
    plt.xlim(lam_0*1e9,lam_1*1e9)
    plt.xlabel(r'wavelength (nm)')
    plt.ylabel(r'$\Delta V_\mathrm{ac}/I_0$ [Hz/($W/\mathrm{cm}^2$)]')
    plt.legend()
    plt.title('scalar light shift')


def frequency_from_wavenumber(wn):
    """
    frequency in Hz from wavenumber in 1/cm^2
    """
    return c*wn*100



import pytest
from lightshifts import LightshiftSolver


def test_lightshift_solver_scalar_lightshift(atom_file, trans_file):
    ls = LightshiftSolver(atom_file, trans_file)
    ls_scalar = ls.scalar_lightshift(lam=670e-9, laser_intensity=2)
    assert pytest.approx(-19.2007990788261)==float(ls_scalar)


def test_lightshift_solver_lightshifts(atom_file, trans_file):
    ls = LightshiftSolver(atom_file, trans_file)
    ls_all = ls.lightshifts(lam=555.7e-9, q=1, mFi=-5/2, laser_intensity=2)
    a = [float(x) for x in ls_all]
    assert pytest.approx([583.819297319377, 16.7205709924352,
        -1.18364597616775])==a


def test_lightshift_solver_total_lightshift(atom_file, trans_file):
    ls = LightshiftSolver(atom_file, trans_file)
    ls_tot = ls.total_lightshift(lam=555.7e-9, q=1, mFi=-5/2, laser_intensity=2)
    assert pytest.approx(599.356222335644)==float(ls_tot)
